#include "Vertex.h"
#include "DecayMode.h"
#include "DecayBuilder.h"
#include "Error.h"
#include <iostream>
#include <sstream>
using namespace std;

// helper function: make a 2-body decay mode using a vertex
TwoBodyDecayMode * TwoBodyDecayFromVertex(Vertex & v, string p) {
  // Add a decay mode using this vertex
  if(v.Count() == 3) {
    try {
      int p_2Spin = Particle::pl[p].TwiceSpin();
      int d1_2Spin, d2_2Spin;
      string d1, d2;
      if(v.Leg(0) == p) {
	d1_2Spin = Particle::pl[v.Leg(1)].TwiceSpin();
	d1 = v.Leg(1);
	d2_2Spin = Particle::pl[v.Leg(2)].TwiceSpin();
	d2 = v.Leg(2);
      }
      else if(v.Leg(1) == p) {
	d1_2Spin = Particle::pl[v.Leg(0)].TwiceSpin();
	d1 = v.Leg(0);
	d2_2Spin = Particle::pl[v.Leg(2)].TwiceSpin();
	d2 = v.Leg(2);
      }
      else if(v.Leg(2) == p) {
	d1_2Spin = Particle::pl[v.Leg(0)].TwiceSpin();
	d1 = v.Leg(0);
	d2_2Spin = Particle::pl[v.Leg(1)].TwiceSpin();
	d2 = v.Leg(1);
      }
      // Build up string to use to report error if necessary
      ostringstream ost;
      ost << "Building decay mode of particle ";
      ost << p << " to particles " << d1 << " and " << d2 << "\n";
      ost << "The spins are not in a configuration that is recognized\n";
      ost << "by this code. It may not be implemented in HELAS, the \n";
      ost << "interface may not exist, or you might have something that\n";
      ost << "is not Lorentz invariant.\n";
      if((d1_2Spin + d2_2Spin + p_2Spin) % 2 != 0) {
	ost << "In particular, you seem to have an odd number of fermions.\n";
      }
      string err = ost.str();
      if(p_2Spin == 0) {
	if(d1_2Spin == 0 && d2_2Spin == 0) {
	  // Add new "S_SS" decay
	  S_SS* dm = new S_SS(p,d1,d2,v);
	  return dm;
	}
	else if(d1_2Spin == 2 && d2_2Spin == 2) {
	  // Add new "S_VV" decay
	  S_VV* dm = new S_VV(p,d1,d2,v); 
	  return dm;
	}
	else if(d1_2Spin == 1 && d2_2Spin == 1) {
	  // Add new "S_FF" decay
	  S_FF* dm = new S_FF(p,d1,d2,v);
	  return dm;
	}
	else if(d1_2Spin == 2 && d2_2Spin == 0) {
	  // Add new "S_VS" decay
	  S_VS* dm = new S_VS(p,d1,d2,v);
	  return dm;
	}
	else if(d1_2Spin == 0 && d2_2Spin == 2) {
	  // Add new "S_VS" decay
	  S_VS* dm = new S_VS(p,d2,d1,v);
	  return dm;
	}
	else {
	  throw UnimplementedError(err);
	}
      } else if(p_2Spin == 1) {
	if(d1_2Spin == 1 && d2_2Spin == 0) {
	  // Add new "F_FS" decay
	  // Need to swap the other fermion
	  d1 = Particle::pl[d1].Antiparticle();
	  F_FS* dm = new F_FS(p,d1,d2,v);
	  return dm;
	} else if(d1_2Spin == 0 && d2_2Spin == 1) {
	  // Add new "F_FS" decay
	  // Need to swap the other fermion
	  d2 = Particle::pl[d2].Antiparticle();
	  F_FS* dm = new F_FS(p,d2,d1,v);
	  return dm;
	} else if(d1_2Spin == 1 && d2_2Spin == 2) {
	  // Add new "F_FV" decay
	  // Need to swap the other fermion
	  d1 = Particle::pl[d1].Antiparticle();
	  F_FV* dm = new F_FV(p,d1,d2,v);
	  return dm;
	} else if(d1_2Spin == 2 && d2_2Spin == 1) {
	  // Add new "F_FV" decay
	  // Need to swap the other fermion
	  d2 = Particle::pl[d2].Antiparticle();
	  F_FV* dm = new F_FV(p,d2,d1,v);
	  return dm;
	} else {
	  throw UnimplementedError(err);
	}
      } else if(p_2Spin == 2) {
	if(d1_2Spin == 1 && d2_2Spin == 1) {
	  // Add new "V_FF" decay
	  V_FF* dm = new V_FF(p,d1,d2,v);
	  return dm;
	} else if(d1_2Spin == 0 && d2_2Spin == 0) {
	  // Add new "V_SS" decay
	  V_SS* dm = new V_SS(p,d1,d2,v);
	  return dm;
	} else if(d1_2Spin == 2 && d2_2Spin == 0) {
	  // Add new "V_VS" decay
          // Need to swap a sign
          d1 = Particle::pl[d1].Antiparticle();
	  V_VS* dm = new V_VS(p,d1,d2,v);
          return dm;
	} else if(d1_2Spin == 0 && d2_2Spin == 2) {
	  // Add new "V_VS" decay
          // Need to swap a sign
          d2 = Particle::pl[d2].Antiparticle();
	  V_VS* dm = new V_VS(p,d2,d1,v);
          return dm;
	} else if(d1_2Spin == 2 && d2_2Spin == 2) {
	  // Add new "V_VV" decay
          // Need to swap a sign (e.g. to get wh+ -> w+ ah in LHT)
          d1 = Particle::pl[d1].Antiparticle();
	  V_VV* dm = new V_VV(p,d1,d2,v);
          return dm;
	} else {
	  throw UnimplementedError(err);
	}
      } else {
    throw UnimplementedError(err);
      }
    }
    catch (UnimplementedError theErr) {
      theErr.report(cerr);
    }
  }
  return NULL;
}

vector<DecayMode *> DecayBuilder::GetTwoBodyDecayModes(string p)
{
  vector<DecayMode *> decayList;
  list<Vertex>::iterator i;
  for(i = Vertex::vl.begin(); i != Vertex::vl.end(); i++) {
    Vertex & v = *i;
    if(v.Multiplicity(p) == 1) {
      if(v.Decays(p)) {
	DecayMode * dm = TwoBodyDecayFromVertex(v,p);
	if(dm != NULL) {
	  decayList.push_back(dm);
	}
      }
    }
  }
  return decayList;
}

vector<DecayMode *> DecayBuilder::GetDecayModes(string p)
{
  // First get the two body modes
  vector<DecayMode *> decayList = GetTwoBodyDecayModes(p);

  // Create a list of 3-body decays to fill
  vector<ThreeBodyDecayMode *> threeDecays;

  // Now look for 3-body decays
  list<Vertex>::iterator i;
  for(i = Vertex::vl.begin(); i != Vertex::vl.end(); i++) {
    Vertex & v = *i;
    if(v.Multiplicity(p) == 1 && 
       (Particle::pl[p].Antiparticle() == p || 
        v.Multiplicity(Particle::pl[p].Antiparticle()) == 0)) {
      // Recursively hunt for decays via off-shell stuff
      if(v.Count() == 3 && !v.Decays(p)) {
	double pmass = Particle::pl[p].Mass();
        // make a 2-body decay object; this will conjugate fields in the proper way.
	TwoBodyDecayMode* dm = TwoBodyDecayFromVertex(v,p);
	if(dm != NULL) {
          // d1 = lighter daughter, d2 = heavier daughter
	  int d1, d2;
	  double lmass, hmass;
	  if(fabs(Particle::pl[dm->ListDaughters()[0]].Mass()) > 
	       fabs(Particle::pl[dm->ListDaughters()[1]].Mass())) {
	    d1 = 1;
	    d2 = 0;
	    lmass = fabs(Particle::pl[dm->ListDaughters()[1]].Mass());
	    hmass = fabs(Particle::pl[dm->ListDaughters()[0]].Mass());
	  } else {
	    d1 = 0;
	    d2 = 1;
	    lmass = fabs(Particle::pl[dm->ListDaughters()[0]].Mass());
	    hmass = fabs(Particle::pl[dm->ListDaughters()[1]].Mass());
	  }
	  if(lmass < pmass) {
	    // It is, so try to decay the heavier one.
	    vector<DecayMode *> chainDecays = 
	      DecayBuilder::GetTwoBodyDecayModes(dm->ListDaughters()[d2]);
  	    vector<DecayMode *>::iterator d;
	    for(d = chainDecays.begin(); d != chainDecays.end(); d++) {
	      if((*d)->SumDaughtersMass() + lmass < pmass) {
	        // Add a new 3-body decay mode or append to
	        // an existing one.
	        vector<string> theDaughters(3);
	        theDaughters[0] = dm->ListDaughters()[d1];
	        theDaughters[1] = (*d)->ListDaughters()[0];
	        theDaughters[2] = (*d)->ListDaughters()[1];
	        vector<ThreeBodyDecayMode *>::iterator d3;
	        bool decayExists = false;
		TwoBodyDecayMode* dm2 = dynamic_cast<TwoBodyDecayMode*>(*d);
		TwoBodyDecayPair pair;
		pair.pr.first = dm;
		pair.pr.second = dm2;
		for(d3 = threeDecays.begin(); d3 != threeDecays.end(); d3++) {
		  if((*d3)->DecayMatch(pair)) {
		    decayExists = true;
		    (*d3)->AddVertexPair(pair);
		    break;
		  }
		}
		if(!decayExists) {
		  ThreeBodyDecayMode *dm3 = new ThreeBodyDecayMode(pair);
		  threeDecays.push_back(dm3);
		}
	      }
	    }
	  }
	  // check if the heavier one is lighter than p.
	  if(hmass < pmass) {
	    // It is, so also try to decay the lighter one.
	    vector<DecayMode *> chainDecays = 
	      DecayBuilder::GetTwoBodyDecayModes(dm->ListDaughters()[d1]);
	    vector<DecayMode *>::iterator d;
	    for(d = chainDecays.begin(); d != chainDecays.end(); d++) {
	      if((*d)->SumDaughtersMass() + hmass < pmass) {
	        // Add a new 3-body decay mode or append to
	        // an existing one.
	        vector<ThreeBodyDecayMode *>::iterator d3;
	        bool decayExists = false;
		TwoBodyDecayMode* dm2 = dynamic_cast<TwoBodyDecayMode*>(*d);
		TwoBodyDecayPair pair;
		pair.pr.first = dm;
		pair.pr.second = dm2;
                cout << "found pair2, dm1: " << dm->Parent() << " -> " << dm->ListDaughters()[0] << ", "
		     << dm->ListDaughters()[1] << "\n" << "       dm2: " <<  dm2->Parent() << " -> " 
                     << dm2->ListDaughters()[0] << ", "
		     << dm2->ListDaughters()[1] << "\n";
		for(d3 = threeDecays.begin(); d3 != threeDecays.end(); d3++) {
		  if((*d3)->DecayMatch(pair)) {
		    decayExists = true;
		    (*d3)->AddVertexPair(pair);
		    break;
		  }
		}
		if(!decayExists) {
		  ThreeBodyDecayMode *dm3 = new ThreeBodyDecayMode(pair);
		  threeDecays.push_back(dm3);
		}
	      }
	    }
	  }
	}
      }
    }
  }
  // Add the three-body decays to the overall list
  vector<ThreeBodyDecayMode *>::iterator d3;
  for(d3 = threeDecays.begin(); d3 != threeDecays.end(); d3++) {
    DecayMode* dm = (DecayMode*)(*d3);
    decayList.push_back(dm);
  }
  
  return decayList;
}
